-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This module interfaces the Presburger API with the toolkit

{-# LANGUAGE CPP#-}
module Assertion.PresburgerValidity
    (checkPresburger)
    where
 
import Data.Integer.Presburger
import Data.Integer.Presburger.Term 
import qualified Data.Integer.Presburger.Form as F
import Common.Absglobaltype
#if __GLASGOW_HASKELL__ >= 610
import Language.Haskell.Interpreter
#else
import System.Eval.Haskell
import System.Plugins.Load
#endif
import System.IO.Unsafe
import Data.Set
import Data.Map as M
import Common.Utils
import Common.PrettyPrinter
import Common.ExpressionTyping
--
-- Returns True if an expression is satisfiable
--
checkPresburger :: Exp -> Bool
checkPresburger exp = check (exp2formula exp)



--
-- Translates a "venocid" expression into a presburger formula
--
exp2formula ::  Exp -> Formula
exp2formula  exp = evaluateExp (translateForm exp) 


--
-- Evaluation function: takes a string which contains a presburger formula
-- and dumps it into a Formula object
--
-- USe System.Eval version when ghc < 6.10, haskell-hint otherwise
--
evaluateExp :: String -> Formula
#if __GLASGOW_HASKELL__ >= 610
evaluateExp s = unsafePerformIO $ hinteval s

--
-- haskell-hint version
-- NB: It is "impossible" not to use unsafePerformIO as
-- System.Eval.eval returns 'IO a' only
-- and runInterpreter requires 'm' s.t. 'MonadCatchIO m'
-- and 'IO a' is such a monad
--
hinteval :: String -> IO Formula
hinteval s = do c <- runInterpreter (formInt s)
                case c of
                  Left err -> error $ printInterpreterError err
                  Right v -> return v
                              
--
-- FIXME: Is it possible not to load these modules all the time?
--
formInt :: String -> Interpreter Formula
formInt s = do setImports ["Prelude","Data.Integer.Presburger.HOAS"]
               c <- interpret s (as :: Data.Integer.Presburger.Formula)
               return c
        
printInterpreterError :: InterpreterError -> String
printInterpreterError e = "\n[haskell-hint-related]\n Ups... " ++ (show e)
#else
-- 
-- System.Eval version
-- 
evaluateExp s =  case unsafePerformIO (eval (s ++ " :: Formula") ["Data.Integer.Presburger"] :: IO (Maybe Formula)) of
                   Nothing -> error $ "Could not evaluate: "++(s)++"\n."
                   Just s -> s
#endif
 

--
-- TODO: boolean variables need to be translated into "integer"
-- ex: v ->  Not(v :=: 0)
-- should be done at parsing time (?)
--


--
-- Translates a "vosenid" expression into a string representing a presburger formula
-- It is assumed that all the variables are quantified!
-- (i.e. LogicAPI always calls mkUnverisal first)
--
translateForm :: Exp -> String
translateForm exp = translateForm_ Empty exp
            
translateForm_ :: Sorting -> Exp -> String
translateForm_ sort BTrue =  "(0 :=: 0)" -- "TRUE"
translateForm_ sort BFalse = "(0 :=: 1)"        -- "FALSE"
translateForm_ sort (BNot e1) = "Not ("++(translateForm_ sort e1)++")"
translateForm_ sort (EBin e1 b e2) = 
    case expTyping sort e1 of
      Just x -> case x of 
                  TInt -> (translateForm_ sort e1)++" "++(translateOp b)++" ("++ (translateForm_ sort e2)++")"
                  TBool -> "("++(translateForBool sort e1 b e2)++")"
                  otherwise -> error $ "At the moment, only boolean and integer variables may be asserted!"
      Nothing -> error $ "[PresburgerValidity] Type error in "++(prPrint e1)++"\nEnvironment:\n"++(prPrint sort)

translateForm_ sort (EAr e1 b e2)  = (translateForm_ sort e1)++" "++(translateArOp b)++" ("++ (translateForm_ sort e2)++")"
translateForm_ sort (EExist xs e) = (translateVar xs " Exists")++" (" ++ (translateForm_ (addVariables sort xs) e)++")"
translateForm_ sort (EForall xs e) = (translateVar xs " Forall")++" (" ++ (translateForm_ (addVariables sort xs) e)++")"
translateForm_ sort (ExpI i) = show i
translateForm_ sort (ExpV v@(Var (Ident s))) = case lookupVariable sort v of
                                                 Just x -> case x of 
                                                             TInt -> s
                                                             TBool -> "Not ("++s++" :=: 0)"
                                                             otherwise -> error $ "At the moment, only boolean and integer variables may be asserted!"
                                                 Nothing -> error $ "Variable "++(show s)++" not found in\n"++(prPrint sort)++"."
translateForm_ sort (ExpS s) = show s
translateForm_ sort s = show s --error ("not supported: "++(prPrint s))


--
-- Translates Venocid boolean operators to presburger ones
--
translateOp :: BinOp -> String
translateOp EOr = ":\\/:" -- 
translateOp EAnd = ":/\\:" --
translateOp EEq = ":=:"
translateOp ENEq = ":/=:"
translateOp EGt = ":>:"
translateOp ELt = ":<:"
translateOp ELte = ":<=:"
translateOp EGte = ":>=:"
translateOp EImpl = ":=>:" 

translateArOp :: ArOp -> String
translateArOp EPlus = "+"
translateArOp EMinus = "-"
translateArOp EDiv = "/"
translateArOp EMul = "*"

translateForBool :: Sorting -> Exp -> BinOp -> Exp -> String
translateForBool sort e1 b e2 = case b of
                EOr -> ne1 ++" :\\/: "++ne2
                EAnd -> ne1 ++" :/\\: "++ ne2
                EImpl ->  ne1 ++" :=>: "++ ne2
                EEq -> ne1++ " :/\\: "++ne2++" :\\/: Not ("++ne1++ " :\\/: "++ne2++")"
                ENEq -> "Not ("++ne1++ " :/\\: "++ne2++" :\\/: Not ("++ne1++ " :\\/: "++ne2++")" ++")"
                EGt -> ne1 ++" :/\\: Not ("++ ne2++")"
                ELt -> ne2 ++" :/\\: Not ("++ ne1++")"
                ELte -> " Not ("++ne2++") :\\/: "++ne1
                EGte -> " Not ("++ne1++") :/\\: "++ne2
    where ne1 = translateForm_ sort e1       
          ne2 = "("++translateForm_ sort e2++")"
  




--
-- translates Venocid variable to presburger ones
--
translateVar :: [VarDecl] -> String -> String
translateVar [] _ = ""
translateVar ((VarDeclaration (Var (Ident s)) t):xs) qu = qu ++ " $ \\ " ++ s ++" -> "++ (translateVar xs qu)





