-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This module provides the functions to generate Haskell code

module CodeGeneration.Generator (generate) where 

import Common.Absglobaltype
import qualified Data.Map as M

type Commu = M.Map Channel [Channel]


--
-- Generates Haskell code of a Process
--
generate :: Process -> String
generate p = mkHeader ++
             mkInit ++
             mkConnect 1 c++
             s ++
             mkEnd
    where (c, s) = genProcess 1 p


--
-- Generates imports
--
mkHeader :: String
mkHeader = "module Main where \n"++
           "import System.IO \n"++
           "import Control.Concurrent \n"++
           "import Control.Concurrent.Chan \n"++
           "import Prelude hiding (catch) \n"++
           "import CodeGeneration.ProcManager \n"++
           "import System.Exit \n"

--
-- Generates main initialisation
-- (i.e. variable to sync the children)
--      
mkInit :: String
mkInit = "main = do { \n" ++
         tab ++ "children <- newMVar [] ;\n"

--
-- Generates end (i.e. synch children)
--
mkEnd :: String
mkEnd = tab ++ "waitForChildren children ; \n"++ 
        tab ++ "putStrLn \"End.\" } "

--
-- Generate code for each concurrent participant
--
mkConcurrent :: [(Commu, String)] -> (Commu, String)
mkConcurrent [] = (M.empty,"")
mkConcurrent ((c, s):xs) = (M.union c c', 
                            tab++
                            "forkChild children ( do { \n" ++ s ++" } );\n"++
                            s')
    where (c', s') = mkConcurrent xs



--
-- Generates communication channels
--
mkConnect :: Int -> Commu -> String
mkConnect ind c = unwords $ map (mkChannels ind) xs
    where xs = map (\(k, v) -> v) (M.toList c)
          mkChannels _ [] = []
          mkChannels ind (c:cs) = printLine ind (chan2haskell c++" <- newChan ;\n")++
                        mkChannels ind cs

--
-- Top level function for concurrent and single participants
--
genProcess :: Int -> Process -> (Commu, String)
genProcess ind (SingleProcess p) = genPrimitive ind p
genProcess ind (ParallelProcess xs) = mkConcurrent $ map (genProcess (ind+1) . (\(ConcP p) -> p)) xs



--
-- Generates code from primtives
--
genPrimitive :: Int -> Primitive -> (Commu, String)
genPrimitive ind x = case x of 
                       PInit c ps cs p -> (M.insert c cs nc, ns)
                           where (nc, ns) = genProcess ind p
                       PJoin c p' cs p  -> (M.insert c cs nc, ns)
                           where (nc, ns) = genProcess ind p
                       PSend c es vds a p -> (nc,
                                              printLine ind ("let "++(genVar vds)++" = "++(genExp es)++"\n") ++
                                              printLine ind ("in\n") ++
                                              printLine (ind+1) ("do {\n") ++
                                              (send (ind+2) c (genVar vds)) ++
                                              ns++
                                              printLine (ind+1) ("\n}\n"))
                           where (nc, ns) = genProcess (ind+2) p
                       PReceive  c vds a p -> (nc, 
                                               debug ind False "begin-receive"++
                                               debug ind True ("[R]waiting on channel "++show c)++
                                               printLine ind ("t' <- " ++ (receive c ))++";\n"++
                                               debug ind True ("[R]received \"++(show t')"++"++\" from "++show c)++
                                               printLine ind ("let "++(genVar vds)++" = read t'::"++(genType vds))++"\n"++
                                               printLine ind ("in\n")++
                                               printLine (ind+1) ("do {\n") ++
                                               printLine (ind+2) (mkMonitor a)++"\n"++
                                               ns++
                                               printLine (ind+1) ("}\n")++
                                               debug ind False "end-receive"
                                              )
                           where (nc, ns) = genProcess (ind+2) p
                       PSelect c a bid la p -> (nc, (send ind c (genLabel bid la) ++ ns))
                           where (nc, ns) = genProcess ind p
                       PBranch c bid brs -> (nbc, 
                                                debug ind False "begin-branch"++
                                                debug ind True ("[B]waiting on channel "++show c)++
                                                printLine ind (brvar bid ++ " <- " ++ receive c) ++";\n"++
                                                debug ind True ("[R]received label [\"++(show "++brvar bid++")++\"] from "++show c)++
                                                printLine ind ("let "++brvar bid++"' = read "++brvar bid++"::String in\n") ++
                                                printLine (ind+1) ("case "++ brvar bid++"' of \n")++
                                                nbs ++
                                                debug ind False "end-branch"
                                            )
                           where (nbc, nbs) = genBranch (ind+2) bid brs
                                 brvar (BId bid) = "brvar"++(show bid)
                       PEnd -> (M.empty, printLine ind "putStrLn  \"end.\"\n")
                       ITE e p1 p2 ->  let (nc1, ns1) = genProcess (ind+2) p1
                                           (nc2, ns2) = genProcess (ind+2) p2
                                       in (M.union nc1 nc2,
                                              debug ind False "begin:if-then-else"++
                                              printLine ind ("if "++ (exp2haskell e))++"\n"++
                                              printLine (ind+1) ("then do {\n") ++
                                              ns1++
                                              printLine (ind+1) "}\n" ++
                                              printLine (ind+1) ("else do {\n") ++
                                              ns2++
                                              printLine (ind+1) "}\n"++
                                              debug ind False "end-if-then-else"
                                          )
                       PRecur v es cs1 vd cs2 p -> (nc,
                                                    debug ind False "begin-recursion"++
                                                    printLine ind ("let "++(genRec v)++" "++genPar vd ++" "++genCs cs2++" = do {\n")++
                                                    -- printLine (ind+1) ("do {\n")++
                                                    ns ++ 
                                                    printLine (ind) "}\n"++
                                                    printLine ind "in\n" ++
                                                    printLine (ind+1) (genRec v++" "++genArg es++" "++(genCs cs1))++"\n"++
                                                    debug ind False "end-recursion"
                                                   )
                           where (nc, ns) = genProcess (ind+2) p
                       PRecCall v es cs -> (M.empty, printLine ind ((genRec v) ++" "++ genArg es ++" "++ genCs cs)++";\n")
                       PDelegS c cs decl a p -> (nc, (send ind c ("\"[delegate "++(show cs)++"]\"")) ++";\n"++ ns)
                           where (nc, ns) = genProcess ind p
                                 -- TODO: use Duplicate for the chan
                       PDelegR c del@(SessPass chs lt p') a p -> (nc, 
                                               debug ind False "begin-receive-deleg"++
                                               debug ind True ("[R]waiting on channel "++show c)++
                                               printLine ind ("t' <- " ++ (receive c ))++";\n"++
                                               debug ind True ("[R]received: \"++(show t')"++"++\" from "++show c)++
                                               ns++
                                               debug ind False "end-receive-deleg"
                                              )
                           where (nc, ns) = genProcess (ind) p
                           

    where genExp = genTuple exp2haskell
          genVar = genTuple var2haskell 
          genRec (Var (Ident v)) = "recvar"++v
          genPar x = genParam var2haskell x
          genCs x = genParam chan2haskell x
          genArg x = genParam exp2haskell x
          genType x = genTuple typ2haskell x
          genChan xs = genTuple (\x -> "Chan") xs

--
-- Label generation
--
genLabel :: BranchId -> Label -> String
genLabel (BId (Ident id)) (Lab (Ident l)) = show(id++l)


--
-- Branches generation
--
genBranch :: Int -> BranchId -> [PBranchElmt] -> (Commu, String)
genBranch ind id xs = foldl (\(nc,ns) -> \(nc',ns') -> (M.union nc nc',ns++ns')) (M.empty,[]) (map (g ind) xs)
    where g ind (PBranchElmt a l p) = (c,
                                        printLine ind ((genLabel id l) ++ " -> " ++ "do {\n")++
                                        s++
                                        printLine ind "}\n"
                                      )
              where (c, s) = genProcess (ind+1) p


--
-- Generates code to bind value to variable
-- (i.e. (v1, ... vn) = (e1,... en)
--
genBindVar :: Int -> [VarDecl] -> [Exp] -> String
genBindVar _ [] [] = []
genBindVar ind (v:vs) (e:es) = printLine ind ((var2haskell v)++" = "++(exp2haskell e)++"\n")
                               ++ genBindVar ind vs es

--
-- Generates code to bind channel to variable
-- (i.e. (v1, ... vn) = (c1,... cn)
--
genBindChan :: Int -> [[Channel]] -> [[Channel]] -> String
genBindChan _ [] [] = ""
genBindChan ind (c:cs) (c':cs') = printLine ind ((genTuple chan2haskell c)++" = "++(genTuple chan2haskell c')++"\n" )
                                  ++ genBindChan ind cs cs'

--
-- Generates list of parameters
--
genParam :: (a -> String) -> [a] -> String
genParam f xs = foldl (\x y-> (y++" "++x)) [] (map f xs)

--
-- Generates tuples
--
genTuple :: (a -> String) -> [a] -> String
genTuple f xs = "("++ (g xs) ++")"
             where g [] = ""
                   g [x] = f x
                   g (x:xs) = (f x) ++ ", " ++ (g xs)
            

--
-- NAME to Haskell
--
chan2haskell (Chan id) = show id 
var2haskell (VarDeclaration (Var v) t) =  show v
typ2haskell (VarDeclaration (Var v) t) = case t of
                                           TInt -> "Int"
                                           TBool -> "Bool"
                                           TString -> "String"
                                           TDate -> "String"
                 
--
-- Expression to Haskell
--
exp2haskell :: Exp -> String
exp2haskell e = "("++translateForm e++")"

--
-- Translate expressions to strings
--
-- to be completed
translateForm :: Exp -> String
translateForm BTrue =  "True"
translateForm BFalse = "False"
translateForm (BNot e1) = "not ("++(translateForm e1)++")"
translateForm (EBin e1 b e2) =(translateOp b)++" ("++ (translateForm e1)++") ("++ (translateForm e2)++")"
translateForm (EAr e1 b e2) =(translateOpA b)++" ("++ (translateForm e1)++") ("++ (translateForm e2)++")"
translateForm (EExist xs e) = (translateForm e)
translateForm (EForall xs e) = (translateForm e)
translateForm (ExpI i) = show i
translateForm (ExpV (Var (Ident s))) = s
translateForm (ExpS s) = show s
translateForm (ExpD d m y) = show (d, m, y)
translateForm s = show s --error ("not supported: "++(prPrint s))

--
-- Operators to strings
--
translateOp :: BinOp -> String
translateOp EOr = "(||)"
translateOp EAnd = "(&&)"
translateOp EEq = "(==)"
translateOp ENEq = "(/=)"
translateOp EGt = "(>)"
translateOp ELt = "(<)"
translateOp ELte = "(<=)"
translateOp EGte = "(>=)"
translateOp EImpl = ":=>:"

translateOpA :: ArOp -> String
translateOpA EPlus = "(+)"
translateOpA EMinus = "(-)"
translateOpA EDiv = "(/)"
translateOpA EMul = "(*)"

--
-- Send actions (i.e. write on channels)
--
send :: Int -> Channel -> String -> String
send ind c s = printLine ind ("writeChan "++(chan2haskell c)++" ( show "++s ++");\n")

--
-- Receive actions (i.e. read on channels)
--
receive :: Channel -> String
receive c = "readChan "++(chan2haskell c)

--
-- Print a line with indentation "ind"
--
printLine :: Int -> String -> String
printLine ind s = (unwords (replicate ind tab)) ++ s

--
-- Tab space
--
tab :: String
tab = " "

--
-- Output some information on the code execution
--
debug :: Int -> Bool -> String -> String
debug ind False s = printLine ind ("-- "++s++"\n")
debug ind True s  = printLine ind ("putStrLn $ \""++s++"\";\n")

--
-- Generates monitor: TO BE IMPLEMENTED
--
mkMonitor :: Assertion -> String
mkMonitor (Assert exp) = "" --"assert "++(translateForm exp)