-- Author: Julien Lange (jl250@le.ac.uk)
-- 03.02.2010

module Global.OneTimeUnfolding where

import Common.Absglobaltype

--
-- A pair consisting of :
--    * The identifier of a recursion variable
--    * Its corresponding DistributedInteractions
--  (name comes from POPL paper's appendix)
-- 
type Scope = (Variable, DistributedInteraction)



--
-- Returns the one-time unfolding of the argument
--
unfoldOneTime :: DistributedInteraction -> DistributedInteraction
unfoldOneTime d = unfold_helper d []



{- ------------------------------------------------------------------- -}
{-              Function Helpers                                       -}
{- ------------------------------------------------------------------- -}

--
-- Takes DistributedInteractions and a Scope
-- and replaces every occurence of a recursion
-- variable by its corresponding "scope"
--
unfold_helper :: DistributedInteraction -> [Scope] -> DistributedInteraction
unfold_helper (ParallelInter []) _ =  ParallelInter []
unfold_helper (ParallelInter ((Conc x):xs)) scope =
    ParallelInter ((Conc (unfold_helper x scope))
                   :
                   (extractInt (unfold_helper (ParallelInter xs) scope)))

unfold_helper (SingleInter x) scope = case x of
            ValPass p1 p2 ch var assert int -> 
                 SingleInter (ValPass p1 p2 ch var assert (unfold_helper int scope))

            Deleg p1 p2 ch del a int ->
                SingleInter (Deleg p1 p2 ch del a (unfold_helper int scope))
             
            LabelPass p1 p2 ch id br ->
               SingleInter (LabelPass p1 p2 ch id (unfold_helper_B br scope))
   
            End -> (SingleInter End)

            Recur var es vs a int -> unfold_helper int ((var, (addEnds var int)):scope)
--Recur var int -> Interactions [Recur var (unfold_helper int ((var, (addEnds var int)):scope))]
            RecCall var es -> (findScope scope var)              
                                                        

--
-- unfold_helper helper (for select/branch nodes)
--                                                    
unfold_helper_B :: [BranchElmt] -> [Scope] -> [BranchElmt]
unfold_helper_B [] _ = []
unfold_helper_B (x:xs) scope = case x of
                                 BranchingElmt assert lab int ->
                                     (BranchingElmt assert lab (unfold_helper int scope):(unfold_helper_B xs scope))


--
-- Given a list of Scope and one variable
-- returns the corresponding Scope
--
findScope :: [Scope] -> Variable -> DistributedInteraction
findScope [] _ = error "No recursion defined"
findScope (x:xs) id 
    | (fst x) == id = snd x
    | (fst x) /= id = findScope xs id


--
-- Returns the List of ConcurInteraction contained
-- in a DistributedInteraction
--      (also used below)
--
extractInt :: DistributedInteraction -> [ConcurInteraction]
extractInt (ParallelInter pi) = pi
extractInt _ = error "No Concurrent Interactions found!"


{- ------------------------------------------------------------------- -}
{-              Replace recursion variables by "End"                   -}
{- ------------------------------------------------------------------- -}                      
-- 
-- Replace every call of Variable by "End"
-- To be called when adding the Scope in the list
-- (otherwise will change every call (i.e. even the first one))
--
addEnds :: Variable -> DistributedInteraction -> DistributedInteraction
addEnds _ (ParallelInter []) =  ParallelInter []
addEnds v (ParallelInter ((Conc x):xs)) = 
    ParallelInter ((Conc (addEnds v x))
                   :
                   (extractInt (addEnds v (ParallelInter xs))))
addEnds v (SingleInter x) = case x of 
                                   RecCall var es -> if var == v
                                                     then SingleInter End
                                                     else SingleInter (RecCall var es)
                                   ValPass p1 p2 ch var asser int -> SingleInter ( ValPass p1 p2 ch var asser (addEnds v int))
                                   Deleg p1 p2 ch del asser int -> SingleInter (Deleg p1 p2 ch del asser (addEnds v int))
                                   LabelPass p1 p2 ch id br -> SingleInter (LabelPass p1 p2 ch id (addEnds_helper v br))
                                   Recur var es vs a int -> SingleInter (Recur var es vs a (addEnds v int))
                                   otherwise -> SingleInter x
    
--
-- addEnds Helper (for select/branch nodes)
--
addEnds_helper :: Variable -> [BranchElmt] -> [BranchElmt]
addEnds_helper _ [] = []
addEnds_helper v ((BranchingElmt a l d):xs) = (BranchingElmt a l newd):(addEnds_helper v xs)
    where newd = addEnds v d


