module Global.Projection where

import Common.Absglobaltype
import Common.PrettyPrinter
import Global.Linearity
import Common.Utils
import Data.Set as Set
import Data.List as L
import Data.Map as M
import Assertion.LogicAPI

{- ------------------------------------------------------------------- -}
{-                               PROJECTION                            -}
{- ------------------------------------------------------------------- -}


--
-- Project a list of global specification and returns a projection map
--
doProjection :: [GlobalSpecification] -> ProjMap
doProjection xs = ProjMap (M.fromList (Prelude.map mkpair xs))
    where mkpair (Global ch d) = (ch, computeProjection d)


--
-- Computes an environment from a GlobalSpecification
-- (computes the projections)
--
addProjection :: [GlobalSpecification] -> Sorting
addProjection gs = Sorting (NameMap M.empty) (ProcessMap M.empty) (doProjection gs)

--
-- Returns the list of projected types (i.e. Participant, LocalType) 
--
computeProjection :: DistributedInteraction -> [Projection]
computeProjection di = 
    if linearityChecker di
    then computeProjection_helper (Set.elems (allParticipants di)) di
    else error "Distributed interactions are not linear!"


--
-- Returns the original list of projection if they are single threaded
-- (i.e. each participant appears only once in the list)
-- throw an error otherwise
--
singleThreading ::  [Projection] -> [Projection]
singleThreading [] = []
singleThreading (pr@(Proj p x):xs) = if not (elem p (participants xs))
                                     then pr:(singleThreading xs)
                                     else errorConcurrency p
                           


{- ------------------------------------------------------------------- -}
{-                         TOP-LEVEL PROJECTION                        -}
{-                        (all the participants)                       -}
{- ------------------------------------------------------------------- -}

--
-- Returns the list of participants (may contain double entries) from a list of projection
--
participants :: [Projection] -> [Participant]
participants [] = []
participants ((Proj p lt):xs) = p:(participants xs)

computeProjection_helper :: [Participant] -> DistributedInteraction-> [Projection]
computeProjection_helper [] _ = []
computeProjection_helper (x:xs) int = (Proj x (participantProjection Empty int x (Assert BTrue) (Set.empty))):(computeProjection_helper xs int)



{- ------------------------------------------------------------------- -}
{-                        CONCURRENT INTERACTION                       -}
{-                         (single participant)                        -}
{- ------------------------------------------------------------------- -}
-- 
-- Returns the projected type of DistributedInteraction for Participant
--   
participantProjection :: Sorting -> DistributedInteraction -> Participant -> Assertion -> Set Variable -> LocalType
participantProjection s (ParallelInter []) _ _ _  = LEnd
participantProjection s (ParallelInter xs) p ca sv = checkSingleThreaded (concurrentProjection s xs p ca sv) p
participantProjection s (SingleInter int) p ca sv = participantProjection_helper s int p ca sv


--
-- Returns the list of local types corresponding to a list of concurrent interactions
-- NB: it should be ensured that every participant appears only
-- once in each element of the list (participants are single threaded)
-- 
concurrentProjection :: Sorting -> [ConcurInteraction] -> Participant -> Assertion -> Set Variable -> [LocalType]
concurrentProjection _ [] _ _ _ = []
concurrentProjection s ((Conc x):xs) p ca sv = (participantProjection s x p ca sv ):(concurrentProjection s xs p ca sv)
               


--check that p is only in ONE of them
checkSingleThreaded :: [LocalType] -> Participant -> LocalType
checkSingleThreaded [] _ = error "[checkSingleThreaded] The list of LocalType is empty"
checkSingleThreaded (x:[]) _ = x
checkSingleThreaded (x:xs) p = checkSingleThreaded_helper (x:xs) p False LEnd

checkSingleThreaded_helper :: [LocalType] -> Participant  -> Bool -> LocalType -> LocalType
checkSingleThreaded_helper [] _ _ current = current
checkSingleThreaded_helper (x:xs) p b current = case x of
                                              LEnd -> checkSingleThreaded_helper xs p b current
                                              otherwise -> if b
                                                           then errorConcurrency p
                                                           else checkSingleThreaded_helper xs p True x




{- ------------------------------------------------------------------- -}
{-                        SINGLE THREADED INTERACTION                  -}
{-                            (single participant)                     -}
{- ------------------------------------------------------------------- -}
--
-- Returns the projected (and asserted) type and for a Single Threaded interaction for Participant
--

-- sv == the set of variable that p knows
participantProjection_helper :: Sorting -> Interaction -> Participant -> Assertion -> Set Variable -> LocalType
participantProjection_helper sort (ValPass p1 p2 ch var (Assert a) rint) p (Assert ca) sv
    | p1 == p   =  Send ch var (simplify $ Assert a) (participantProjection nsort rint p gproj newvars)
    | p2 == p   =  Receive ch var (simplify $ Assert (existvExt sort newvars (mkAnd a ca))) (participantProjection nsort rint p gproj newvars)
    | otherwise = participantProjection nsort rint p gproj sv
    where newvars = upVars sv var
          gproj   = simplify $ Assert (mkAnd a ca)
          nsort = addVariables sort var
participantProjection_helper sort (Deleg p1 p2 ch del (Assert a) rint) p (Assert ca) sv
    | p1 == p   =  LDelegS ch del (simplify $ Assert a) (participantProjection sort rint p gproj sv)
    | p2 == p   =  LDelegR ch del (simplify $ Assert (existvExt sort sv (mkAnd a ca))) (participantProjection sort rint p gproj sv)
    | otherwise = participantProjection sort rint p gproj sv
    where gproj = simplify $ Assert (mkAnd a ca)
participantProjection_helper sort (LabelPass p1 p2 ch id xs@(br:brs)) p ca sv
    | p1 == p = Select ch id (labelProjectionSelect sort (br:brs) p ca sv)
    | p2 == p = Branch ch id (labelProjectionBranch sort (br:brs) p ca sv)
    | otherwise = branchProjection sort xs p ca sv
                  -- FIXME: add UNION of projections in otehrwise case
participantProjection_helper sort (Recur var es vs as@(Assert a) dint) p (Assert ca) sv = 
    if proj == (LRecur var es nvar as (LRecCall var es))
    then LEnd -- tail is simplified to ".end" in case projection returns "mu t.t"
    else if isRecursive var sbproj -- one needs also to avoid "mu t.X" 
         then proj                 -- where t doesn't appeart in X
         else sbproj
    where proj = LRecur var (simplExp sv es) nvar (simplify $ Assert (existvExt sort newvars a)) sbproj
          sbproj = participantProjection nsort dint p (Assert (EBin ca EAnd a)) newvars
          -- gproj = simplify $ Assert (EBin a EAnd ca)
          nvar = (varLoc2Var p vs) -- TODO: remove unused variables
          newvars = upVars sv nvar
          nsort = addVariables sort (L.map (\(VarDeclLoc v l) -> v) vs)
participantProjection_helper _ (RecCall var es) _ _ sv = LRecCall var (simplExp sv es)
participantProjection_helper _ End p _ _ = LEnd 
participantProjection_helper _ int _ _ _ = error $ "[Projection] unknown construct: "++(show int)




--
-- returns True if the (process) Variable is used, False otherwise
--
isRecursive :: Variable -> LocalType -> Bool
isRecursive v x = case x of 
                         (Send c1 t1 a1 lt1) -> isRecursive v lt1
                         (LDelegS c de as lt1) -> isRecursive v lt1
                         (Receive c1 t1 a1 lt1) -> isRecursive v lt1
                         (LDelegR c de as lt1) -> isRecursive v lt1
                         (Branch c1 id1 ls1) -> branch ls1
                         (Select c1 id1 ls1) -> branch ls1
                         (LRecur v1 e1 vd1 a1 lt1) -> isRecursive v lt1
                         (LRecCall v1 e1) -> (v == v)
                         lt -> False
    where branch [] = False
          branch xs = foldl (||) False (L.map ((isRecursive v) . (\(LabelElmt a l lt) -> lt)) xs)




--
-- Helpers
--
-- TODO: should remove all the expressions which don't refer to a variable in Set
simplExp :: Set Variable -> [Exp] -> [Exp]
simplExp sv exp = exp

vExt :: Exp -> Set Variable -> [Variable]
vExt a sv = Set.toList $ Set.difference (variableSet a) sv






-- /!\
-- TODO: add a Sorting in expression here types of variable are HARDCODED
--
existvExt :: Sorting -> Set Variable -> Exp -> Exp
existvExt s v exp = case list of 
                    [] -> exp
                    v -> EExist (typelist list) exp
    where list = vExt exp v
          typelist [] = []
          typelist (x:xs) = case lookupVariable s x of
                              Nothing -> error $ "[Projection] Variable "++(prPrint x)++" not found."
                              Just t -> (VarDeclaration x t) : (typelist xs)
                         
--
-- Update set of current variables
--
upVars :: Set Variable -> [VarDecl] -> Set Variable
upVars sv [] = sv
upVars sv ((VarDeclaration v t):xs) = Set.union (upVars sv xs) (Set.singleton v)

--
-- Returns a list of types from variable declarations
--
declTyping :: [VarDecl] -> [Typ] -- type of a list of variable declaration (ex Receive)
declTyping [] = []
declTyping ((VarDeclaration v t):xs) = t:(declTyping xs)

--
-- Returns a list of variable declarations from (declaration, localtion)
--
varLoc2Var :: Participant -> [VarDeclLoc] -> [VarDecl]
varLoc2Var _ [] = []
varLoc2Var p ((VarDeclLoc v (Loc p1 p2)):vs) =
    if p1 == p || p2 == p then v:(varLoc2Var p vs)
    else varLoc2Var p vs


--
-- Project a select construct for a Participant
--
labelProjectionSelect :: Sorting -> [BranchElmt] -> Participant -> Assertion -> Set Variable -> [LabelElmt] 
labelProjectionSelect _ [] _ _ _ = []
labelProjectionSelect sort ((BranchingElmt (Assert a) l dint):xs) p (Assert ca) sv = 
    (LabelElmt (simplify $ Assert a) l (participantProjection sort dint p gproj sv))
    :
    (labelProjectionSelect sort xs p (Assert ca) sv)
    where gproj = Assert (EBin a EAnd ca)


--
-- Project a branch construct for a Participant
--
labelProjectionBranch :: Sorting -> [BranchElmt] -> Participant -> Assertion -> Set Variable -> [LabelElmt] 
labelProjectionBranch _ [] _ _ _ = []
labelProjectionBranch sort ((BranchingElmt (Assert a) l dint):xs) p (Assert ca) sv =  
    (LabelElmt (simplify $ Assert (existvExt sort sv (mkAnd a ca))) l (participantProjection sort dint p gproj sv))
    :
    (labelProjectionBranch sort xs p (Assert ca) sv)
    where gproj = Assert (mkAnd a ca)


--
-- check that Gj\p is the same forall j
--

branchProjection :: Sorting -> [BranchElmt] -> Participant -> Assertion -> Set Variable -> LocalType
branchProjection sort xs p a sv = if length grped == 1
                             then head $ head grped
                             else errorBranch p (head $ head grped) (last $ last grped)
    where projlist = L.filter (equalRecVar) (L.map (\(BranchingElmt _ _ x) -> participantProjection sort x p a sv) xs)
          grped = L.group projlist -- FIXME: find something better that group to test if they're all the same

equalRecVar :: LocalType -> Bool
equalRecVar (LRecur _  _ _ _ (LRecCall _ _)) = False
equalRecVar (LRecur _ _ _ _ LEnd) = False
equalRecVar (LRecCall _ _) = False
equalRecVar _ = True

