-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This modules provides function to ensure a global description
-- is WELL-ASSERTED

module Global.WellAsserted (wellAsserted) where 

import Assertion.LogicAPI
import Common.Absglobaltype
import Common.PrettyPrinter
import Data.Map as Map
import Data.List as List
import Data.Set as Set
import Common.Utils


--
-- Returns True if the description is well-asserted
--
wellAsserted :: DistributedInteraction -> Bool
wellAsserted di = gSat di (Assert BTrue) Global.WellAsserted.empty


--
-- Corresponds to gSat in "A Theory of DoC..."
--
gSat :: DistributedInteraction -> Assertion -> Envi -> Bool
gSat  _                 (Assert BFalse) _  = True
gSat (SingleInter int)  a               e  = gSatInteraction int a e
gSat (ParallelInter xs) a               e  = foldl (&&) True (List.map (\(Conc x) -> gSat x a e) xs)
 
--
-- Environment mapping process variables to (recursion) assertion and participant to (located) variable decl
--
type ProcEnvi = Map Variable (Exp, [Variable], [(Typ, Location)])
type VarEnvi = Map Variable (Typ, Location)

data Envi = Envi ProcEnvi VarEnvi

--
-- Functions for Envi
--
empty :: Envi
empty = Envi Map.empty Map.empty

lookupProc :: Variable -> Envi -> Maybe (Exp, [Variable], [(Typ, Location)])
lookupProc var (Envi pm _) =  Map.lookup var pm

insertProc :: Variable -> (Exp, [Variable], [(Typ, Location)]) -> Envi -> Envi
insertProc v ev (Envi pm vm) = Envi (Map.insert v ev pm) vm

lookupVar :: Variable -> Envi -> Maybe (Typ, Location)
lookupVar p (Envi _ (vm)) = Map.lookup p vm

insertVar :: Variable -> Typ -> Location -> Envi -> Envi
insertVar v t l e@(Envi pm (vm)) = Envi pm ((Map.insert v (t, l) vm))

insertVars :: Envi -> [VarDecl] -> Location -> Envi
insertVars e [] _ = e
insertVars e ((VarDeclaration v t):vs) l = insertVar v t l (insertVars e vs l)
                  
upRecurEnvi :: Envi -> [VarDeclLoc] -> Envi
upRecurEnvi  e [] = e
upRecurEnvi e ((VarDeclLoc (VarDeclaration v t) l):xs) = insertVar v t l (upRecurEnvi e xs) 


sorting :: Envi -> Sorting
sorting (Envi _ vm) = Sorting (NameMap vm') emptyProcessMap emptyProjMap
    where vm' = Map.fromList $ Prelude.map (\(v, (t, l)) -> (v, t)) (Map.toList vm)


--
-- Returns types and location from a declaration with location list
--
location :: [VarDeclLoc] -> [(Typ, Location)]
location [] = []
location ((VarDeclLoc (VarDeclaration v t) l):vs) = (t,l):location vs



--
-- gSat for prexifes
--
gSatInteraction :: Interaction -> Assertion -> Envi -> Bool
gSatInteraction (ValPass p1 p2 _ vs (Assert a') int) (Assert a) e = 
    if (historySensitivity p1 e [a'] vars)  &&
       checkSatis sort impl
    then gSat int (Assert (EBin a EAnd a')) newenvi
    else error $ "[VP]The expresssion " ++ (prPrint impl)++" is not satisfiable."
    where impl = (EBin a EImpl (EExist vs a'))
          vars = variable vs
          newenvi = insertVars e vs (Loc p1 p2)
          sort = sorting e

gSatInteraction (Deleg p1 p2 c deleg (Assert a') int) (Assert a) e =
    if (historySensitivity p1 e [a'] Set.empty)  &&
       checkSatis sort impl
    then gSat int (Assert (EBin a EAnd a')) e
    else error $ "[Deleg]The expresssion " ++ (prPrint impl)++" is not satisfiable."
    where impl = (EBin a EImpl a')
          sort = sorting e
gSatInteraction (LabelPass p1 p2 _ _ bs) a e = gSatBranch p1 bs a e
gSatInteraction End a e = True
gSatInteraction (Recur v es vs (Assert a') int) (Assert a) envi = 
    if checkRecurVar newenvi a' &&
       checkSatis sort (EBin a EImpl a'')
    -- TODO: might have to checked match between es and vs also?
    then gSat int (Assert (EBin a EAnd a')) newenvi
    else error $ "[Rec]The expresssion " ++ (prPrint (EBin a EImpl a''))++" is not satisfiable."
    where a'' = substitute (es, vars) a'
          vars =  List.map (\(VarDeclLoc (VarDeclaration v t) loc) -> v) vs
          loc = location vs
          newenvi = insertProc v (a',vars, loc) (upRecurEnvi envi vs)
          sort = sorting newenvi
gSatInteraction (RecCall v es) (Assert a) envi = 
    case lookupProc v envi of
      Nothing -> error ("Process variable "++(prPrint v)++" not found.")
      Just (a',vs,tl) -> if checkSatis sort (EBin a EImpl (substitute (es, vs) a'))
                           then True
                           else error $ "[RecCall]The expresssion " ++ 
                                (prPrint  (EBin a EImpl (substitute (es, vs) a')))++
                                " is not satisfiable."
    where sort = sorting envi
gSatInteraction i a e = error $ "[gSat: unknown construct]: "++(show i) ++ "\n" ++ (show a)

--
-- gSat for branches
--
gSatBranch :: Participant -> [BranchElmt] -> Assertion -> Envi -> Bool
gSatBranch _ [] _ _ = True
gSatBranch p xs (Assert a) e = if historySensitivity p e assertList Set.empty && 
                               checkSatis sort impl
                               then foldl (&&) True (List.map restassert xs)
                               else error $ "[LB]The expresssion " ++ (prPrint impl)++" is not satisfiable."
    where condassert (BranchingElmt (Assert a') l int) = a'
          restassert (BranchingElmt (Assert a') l int) = gSat int (Assert (EBin a EAnd a')) e
          impl = (EBin a EImpl (foldl (\x y -> EBin x EOr y) BFalse assertList))
          assertList = List.map condassert xs
          sort = sorting e
--
-- Check that recursion variables have been declared
--                                         
checkRecurVar :: Envi -> Exp -> Bool
checkRecurVar (Envi pm (vm)) exp = if sub
                                   then sub
                                   else error $"[checkRecurVar] The variables: "++(show vexp)++
                                           "\n have all been delcared in Envi: "++(show venvi)
    where vexp = (variableSet exp)
          venvi = (Map.keysSet vm)
          sub = Set.isSubsetOf vexp venvi


--
-- Check for history sensitivity
-- i.e. do participants assert only variable they know
--
historySensitivity :: Participant -> Envi -> [Exp] -> Set Variable -> Bool
historySensitivity p envi exp var = historySensitivity_helper p envi diff
    where diff = Set.toList $ Set.difference (Set.unions (List.map variableSet exp)) var -- var(A)\v

historySensitivity_helper :: Participant -> Envi -> [Variable] -> Bool
historySensitivity_helper _ _ [] = True
historySensitivity_helper p envi (x:xs) = case lookupVar x envi of
                                           Nothing -> error $ "Variable "++(prPrint x)++" is not declared."
                                           Just (t,(Loc p1 p2)) -> if p == p1 || p == p2
                                                                   then historySensitivity_helper p envi xs
                                                                   else error $ "Participant "++(prPrint p)
                                                                            ++" doesn't know variable "++(prPrint x)
                                                                            ++", known by "++(prPrint p1)++" and "++(prPrint p2)++"."