-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This module provides a function to type 
-- binary/dyadic sessions
-- (i.e. no need for projections here)
--
--
-- (NB: in binary session, there should be only one shared channel
-- and there's a bijection shared-name <-> public name)
--


module Local.BinaryTyping 
    (binarySessionTyping)
    where

import Local.TypingSystem
import Common.Absglobaltype
import Common.Typing
import Common.PrettyPrinter
import Data.List as L


--
-- Types a process
--
binarySessionTyping :: Process -> Typing
binarySessionTyping p = sessionTypingConc compatible compose Empty p

--
-- Duality
--
dual :: LocalType -> LocalType
dual (Send ch1 ts1 assert1 lt1) = Receive ch1 ts1 assert1 (dual lt1)
dual (Receive ch1 ts1 assert1 lt1) = Send ch1 ts1 assert1 (dual lt1)
dual (Branch ch1 id1 ls1) = Select ch1 id1 (dual_helper ls1)
dual (Select ch1 id1 ls1) = Branch ch1 id1 (dual_helper ls1)
dual (LRecur v e vd a lt) = LRecur v e vd a (dual lt)
dual (LRecCall v e) = LRecCall v e
dual LEnd = LEnd

dual_helper :: [LabelElmt] -> [LabelElmt]
dual_helper [] = []
dual_helper ((LabelElmt ass1 l1 lt1):xs) = (LabelElmt ass1 l1 (dual lt1)):(dual_helper xs)


--
-- Returns True if the two types are compatible (see definition in 
-- "Language primitives and type discipline for structured
-- communication-based programming"
--
compatible :: Typing -> Typing -> Bool
compatible ((c1,p1,cs1,lt1):xs) [] = True
compatible []  ((c2,p2,cs2,lt2):ys) = True
compatible [] [] = True
compatible (l1@(c1,p1,cs1,lt1):xs) (l2@(c2,p2,cs2,lt2):ys) = 
    if (c1 == c2) && (cs1 == cs2)
    then if (ltEq lt1 (dual lt2))
         then (compatible xs ys)
         else  error ("Types "++(prPrint lt1)++"\nand\n "++(prPrint lt2)++"\nare not compatible.")
    else (compatible (l1:xs) ys) && (compatible xs (l2:ys))

--
-- Compose two types
--
-- NB: We assume typings to be compatible
--
compose :: Typing -> Typing -> Typing
compose xs ys = compose_helper [] xs ys

compose_helper :: [(Channel, [Channel])] -> Typing -> Typing -> Typing
compose_helper _ [] xs = xs
compose_helper _ xs [] = xs
compose_helper chs (l1@(c1,p1,cs1,lt1):xs) y@(l2@(c2,p2,cs2,lt2):ys)
    | (c1 == c2) && (cs1 == cs2) 
        = (mergeType l1 y):(compose_helper ((c1,cs1):chs) xs ys)
compose_helper chs (l1@(c1,p1,cs1,lt1):xs) y@(l2@(c2,p2,cs2,lt2):ys)
    | (c1,cs1) `notElem` chs 
        = (mergeType l1 y):(compose_helper ((c1,cs1):chs) xs (l2:ys))
compose_helper chs x@(l1@(c1,p1,cs1,lt1):xs) (l2@(c2,p2,cs2,lt2):ys)
    | (c2,cs2) `notElem` chs 
        = (mergeType l2 x):(compose_helper ((c2,cs2):chs) (l1:xs) ys)
compose_helper chs (l1@(c1,p1,cs1,lt1):xs) y@(l2@(c2,p2,cs2,lt2):ys)
    | ((c2,cs2) `notElem` chs) && ((c1,cs1) `notElem` chs) 
        = l1:(l2:(compose_helper ((c1,cs1):((c2,cs2):chs)) xs ys))

--
-- Returns type _|_ when a type as a counterpart, the type otherwise
-- 
mergeType :: (Channel,Participant,[Channel],LocalType) -> Typing -> (Channel,Participant,[Channel],LocalType)
mergeType l1@(c1, p1, cs1, lt1) y@((c2, p2, cs2, lt2):ys) =
    if isPresent l1 y
    then (c1, p1, cs1, Bottom)
    else l1


--
-- Returns True if the shared name<->private channels have already been typed in Typing
--
isPresent :: (Channel,Participant,[Channel],LocalType) -> Typing -> Bool
isPresent _ [] = False
isPresent l@(c, p, ps, lt) ((c1, p1, ps1, lt1):xs) =
    if (c == c1) && (ps == ps1)
    then  True
    else isPresent l xs
         
         
--
-- LocalType equality module assertions
--
ltEq :: LocalType -> LocalType -> Bool
ltEq (Send c1 t1 a1 lt1)     (Send c2 t2 a2 lt2) =
  (c1 == c2) && (t1 == t2) && (ltEq lt1 lt2)

ltEq (LDelegS c1 d1@(SessPass cs1' lt1' p1) a1 lt1) (LDelegS c2 d2@(SessPass cs2' lt2' p2) a2 lt2) =
  (c1 == c2) && (cs1' == cs2') && (p1 == p2) && (ltEq lt1' lt2') && (ltEq lt1  lt2)

ltEq (Receive c1 t1 a1 lt1)     (Receive c2 t2 a2 lt2) =
  (c1 == c2) && (t1 == t2) && (ltEq  lt1  lt2)
  

ltEq (LDelegR c1 d1@(SessPass cs1' lt1' p1) a1 lt1) (LDelegR c2 d2@(SessPass cs2' lt2' p2) a2 lt2) =
    (c1 == c2) && (cs1' == cs2') && (p1 == p2) && (ltEq  lt2' lt2') && (ltEq lt1 lt2)

ltEq (Branch c1 id1 ls1)     (Branch c2 id2 ls2) = 
    (c1 == c2) && (id1 == id2) && (foldl (&&) True (L.map (\(x,y) -> f x y) lzip))
      where lzip = zip ls1 ls2
            f (LabelElmt a l lt) (LabelElmt a' l' lt') = a == a' && l == l' && (ltEq lt lt')
            f _ _ = False

ltEq (Select c1 id1 ls1)        (Select c2 id2 ls2) =
    (c1 == c2) && (id1 == id2) && (foldl (&&) True (L.map (\(x,y) -> f x y) lzip))
      where lzip = zip ls1 ls2
            f (LabelElmt a l lt) (LabelElmt a' l' lt') = l == l' && ltEq lt lt' 
            f _ _ = False
    
ltEq (LRecur v1 e1 vd1 a1 lt1)  (LRecur v2 e2 vd2 a2 lt2) =
    (v1 == v2) && (e1 == e2) && (vd1 == vd2) && (ltEq lt1 lt2)
    
ltEq (LRecCall v1 e1)      (LRecCall v2 e2) =  (v1 == v2)
                                               -- should test the type of the rec parameters
                                               -- here but it done at the recursion definition
                                               -- in TypingSystem && here.
ltEq LEnd                  LEnd                          = True
ltEq Bottom                Bottom                        = True
ltEq LEnd                  (LRecur _ _ _ _ LEnd)         = True
ltEq (LRecur _ _ _ _ LEnd) LEnd                          = True
ltEq t1                    t2                       =
    error $ "[Binary Typing]Two types are incompatible: \n*"++(prPrint t1)++"\n*"++(prPrint t2)


--
-- Test if two typings are equal, without testing participants identifiers
--
binEq :: Typing -> Typing -> Bool
binEq [] [] = True
binEq [] _ = False
binEq _ [] = False
binEq ((ch,p,chs,lt):xs) ((ch2,p2,chs2,lt2):ys) = (ch == ch2) && 
                                                  (chs == chs2) &&
                                                  (lt == lt) && (binEq xs ys)
                                                  